/******************************************************************************
flt_PNG.cpp : Defines the entry point for the DLL application.

Copyright John Paul Chacha. All Rights Reserved.

This source code file, which has been provided by John Paul Chacha as part of 
his software product for use only by licensed users of that product, includes 
proprietary information of John Paul Chacha.

USE OF THIS SOFTWARE IS GOVERNED BY THE TERMS AND CONDITIONS OF THE LICENSE 
AGREEMENT FURNISHED WITH THE PRODUCT.

IN PARTICULAR, JOHN PAUL CHACHA SHALL BE FREE FROM ANY CLAIMS OR LIABILITIES 
ARISING OUT OF THE USE OR MISUSE OF THIS FILE AND/OR ITS CONTENTS.
******************************************************************************/
#include "../common/system.h"
#include "../common/plugin.h"
#include "../common/suite_info.h"
#include "../common/win_util.h"
#include "../common/str_util.h"
#include "../lib/libpng/png.h"
#include "../lib/libpng/pngstruct.h"
#include "../lib/zlib/zlib.h"
#include "resource.h"
#include "flt_PNG.h"


//=============================================================================
#define PLUGIN_NAME		L"PNG File-format Plug-in"
#define PLUGIN_AUTHOR	SUITE_CREATOR
#define PLUGIN_VERSION	SUITE_VER_INT32

#define MY_STORE_NAME	L"flt_PNG.dll"
#define MY_STORE_UUID	(*((LPDWORD)("_PNG")))

//-----------------------------------------------------------------------------
#define SETTINGS_VERSION ((1<<16)|(1))
struct plugin_config
{
	unsigned long	version;
	bool			always_apply_bKGD;
};

//-----------------------------------------------------------------------------
HINSTANCE api_inst;
plugin_config settings={0,};

LRESULT CALLBACK msg_dlg_about (HWND hDlg,UINT msg,WPARAM wpar,LPARAM lpar);
LRESULT CALLBACK msg_dlg_config(HWND hDlg,UINT msg,WPARAM wpar,LPARAM lpar);

//=============================================================================
//=============================================================================



/******************************************************************************
WINDOWS (Win32) ENTRY POINT FOR THE DYNAMIC-LINK LIBRARY
******************************************************************************/
BOOL APIENTRY DllMain(
	HANDLE hModule,DWORD ul_reason_for_call,LPVOID lpReserved)
{
	//-------------------------------------------------------------------------
	switch(ul_reason_for_call)
	{
	case DLL_PROCESS_ATTACH:
		api_inst=(HINSTANCE)hModule;
		break;
	case DLL_THREAD_ATTACH:
		break;
	case DLL_THREAD_DETACH:
		break;
	case DLL_PROCESS_DETACH:
		break;
	}
	//-------------------------------------------------------------------------
	UNREFERENCED_PARAMETER(lpReserved);
	return TRUE;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/******************************************************************************
This function is called to query plugin's name and type
******************************************************************************/
int plg_GetInfo(plg_INFO* info)
{
	info->api_type=PLUGIN_APITYPE_FILE;
	info->api_version=PLUGIN_INTERFACE_VERSION;
	info->plugin_version=PLUGIN_VERSION;
	wcsncpy(info->plugin_name,PLUGIN_NAME,63);
	info->plugin_name[63]=0;
	wcsncpy(info->plugin_author,PLUGIN_AUTHOR,63);
	info->plugin_author[63]=0;
	return 1;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/******************************************************************************
This function is called to display plugin dialog boxes
******************************************************************************/
int plg_ShowDialog(plg_DIALOG* data)
{
	int i;

	if(!data)
	{
		return PLUGIN_ERR_BAD_PARAM;
	}
	memset(&settings,0,sizeof(plugin_config));
	if(data->pi_StateStore)
	{
		if(data->pi_StateStore->version>=PI_STATESTORE_VERSION)
		{
			data->pi_StateStore->perm_read(
				MY_STORE_NAME,&settings,sizeof(plugin_config));
			if(settings.version!=SETTINGS_VERSION)
			{
				memset(&settings,0,sizeof(plugin_config));
				settings.version=SETTINGS_VERSION;
			}
		}
	}

	switch(data->dialog_id)
	{
	case PLG_DIALOG_ABOUT:
		DialogBoxParam(api_inst,(LPCTSTR)IDD_ABOUT,data->hwnd,
			(DLGPROC)msg_dlg_about,0);
		return PLUGIN_OKAY;
	case PLG_DIALOG_CONFIG:
		i=DialogBoxParam(api_inst,(LPCTSTR)IDD_CONFIG,data->hwnd,
			(DLGPROC)msg_dlg_config,0);
		if(i>0)
		{
			settings.version=SETTINGS_VERSION;
			i=data->pi_StateStore->perm_write(
				MY_STORE_NAME,&settings,sizeof(plugin_config));
			if(!i)
			{
				MessageBox(data->hwnd,
					L"Error saving configuration data",
					L"Internal Error",MB_OK|MB_ICONERROR);
				break;
			}
		}
		return 1;
	case PLG_DIALOG_HELP:
		return PLUGIN_ERR_NO_SUPPORT;
	}
	return PLUGIN_ERR_BAD_PARAM;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/******************************************************************************
I'm sure the nice fellows at ->PNG had a good reason for not
using the time-tested 'magic-number' method.
******************************************************************************/
#ifndef png_jmpbuf
#  define png_jmpbuf(png_ptr) ((png_ptr)->jmpbuf)
#endif

int check_if_png(char *file_name, FILE **fp)
{
   char buf[4];
   if((*fp = fopen(file_name, "rb"))==NULL) return 0;
   if(fread(buf,1,4,*fp)!=4) return 0;
   return(!png_sig_cmp((PBYTE)buf,(png_size_t)0,4));
}

wchar_t str_error_list[4096];
void my_png_error(png_structp ppng, png_const_charp error)
{
	wchar_t werror[1024];
	mbstowcs(werror,error,1024);
	wcscat(str_error_list,werror);
	wcscat(str_error_list,L"\r\n");
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/******************************************************************************
This is the plugin DLL function that Chasys Draw IES calls to check the load 
operation support level for a specified file extension
******************************************************************************/
int flt_CheckLoad(wchar_t* extension)
{
	if(_wcsicmp(extension,L".png")==0) return 100;
	return 0;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/******************************************************************************
This is the plugin DLL function that Chasys Draw IES calls to check the save 
operation support level for a specified file extension
******************************************************************************/
int flt_CheckSave(wchar_t* extension)
{
	if(_wcsicmp(extension,L".png")==0) return 100;
	return 0;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/******************************************************************************
This is the plugin DLL function that Chasys Draw IES calls to load images
******************************************************************************/
int flt_LoadImage(flt_IMAGE_L* data)
{
	int bit_depth,color_type,interlace_type,last_error;
	unsigned int row,sig_read=0;
	png_uint_32 width,height;
	png_color_16 my_bKGD;
	png_color_16 *img_bKGD;
	png_structp png_ptr;
	png_infop info_ptr;
	png_bytep row_pointers[64*1024];
	FILE *fp;

	//-------------------------------------------------------------------------
	if(!data)
	{
		return PLUGIN_ERR_BAD_HOST;
	}
	if(!data->pi_BasicImage)
	{
		return PLUGIN_ERR_BAD_HOST;
	}
	if(data->pi_BasicImage->version<PI_BASICIMAGE_VERSION)
	{
		return PLUGIN_ERR_BAD_HOST;
	}

	//-------------------------------------------------------------------------
	memset(&settings,0,sizeof(plugin_config));
	if(data->pi_StateStore)
	{
		if(data->pi_StateStore->version>=PI_STATESTORE_VERSION)
		{
			data->pi_StateStore->perm_read(
				MY_STORE_NAME,&settings,sizeof(plugin_config));
			if(settings.version!=SETTINGS_VERSION)
			{
				memset(&settings,0,sizeof(plugin_config));
				settings.version=SETTINGS_VERSION;
			}
		}
	}

	//-------------------------------------------------------------------------
	//create and initialize library structures
	png_ptr=png_create_read_struct(PNG_LIBPNG_VER_STRING,0,0,0);
	if(!png_ptr)
	{
		return PLUGIN_ERR_INTERNAL;
	}

	png_ptr->error_fn=my_png_error;
	str_error_list[0]=0;
	info_ptr=png_create_info_struct(png_ptr);
	if(!info_ptr)
	{
		png_destroy_read_struct(&png_ptr,&info_ptr,NULL);
		return PLUGIN_ERR_INTERNAL;
	}

	last_error=PLUGIN_ERR_INTERNAL;
	if(setjmp(png_jmpbuf(png_ptr)))
	{
		goto err_internal;
	}

	//-------------------------------------------------------------------------
	//open the file and read the header
	fp=_wfopen(data->file,L"rb");
	if(!fp)
	{
		return PLUGIN_ERR_FILE_OPEN_R;
	}

	last_error=PLUGIN_ERR_NO_SUPPORT;
	png_init_io(png_ptr,fp);
	png_set_sig_bytes(png_ptr,sig_read);
	png_read_info(png_ptr,info_ptr);
	png_get_IHDR(png_ptr,info_ptr,&width,&height,&bit_depth,&color_type,
		&interlace_type,NULL,NULL);

	//-------------------------------------------------------------------------
	//set up stuff
	last_error=PLUGIN_ERR_INTERNAL;

	png_set_swap(png_ptr);//swap bytes of 16 bit files to small endian
#ifdef PNG_READ_SCALE_16_TO_8_SUPPORTED
	png_set_scale_16(png_ptr);
#else
	png_set_strip_16(png_ptr);
#endif

	png_set_packing(png_ptr);//expand bpp=1/2/4 to full bytes
	png_set_invert_mono(png_ptr);
	if(color_type==PNG_COLOR_TYPE_PALETTE)
	{
		png_set_palette_to_rgb(png_ptr);
	}
	if((color_type==PNG_COLOR_TYPE_GRAY)||(color_type==PNG_COLOR_TYPE_GRAY_ALPHA))
	{
		if(bit_depth<8)
		{
			png_set_expand_gray_1_2_4_to_8(png_ptr);
		}
		png_set_gray_to_rgb(png_ptr);
	}
	if(png_get_valid(png_ptr,info_ptr,PNG_INFO_tRNS))
	{
		png_set_tRNS_to_alpha(png_ptr);
	}
	if(color_type&PNG_COLOR_MASK_COLOR)
	{
		png_set_bgr(png_ptr);//flip RGB/RGBA to BGR/BGRA
	}
	png_set_invert_alpha(png_ptr);
	png_set_filler(png_ptr,0,PNG_FILLER_AFTER);//filler for non-alpha files

	//background color
	my_bKGD.red  =0xffff;
	my_bKGD.green=0xffff;
	my_bKGD.blue =0xffff;
	my_bKGD.gray =0xffff;
	my_bKGD.index=0;
	if( (settings.always_apply_bKGD)&&
		(png_get_bKGD(png_ptr,info_ptr,&img_bKGD)))
	{
		png_set_background(png_ptr,img_bKGD,PNG_BACKGROUND_GAMMA_FILE,1,1);
	}
	else
	{
		png_set_background(png_ptr,&my_bKGD,PNG_BACKGROUND_GAMMA_UNKNOWN,0,1);
	}

	//-------------------------------------------------------------------------
	//optional call to gamma correct and add the background to the palette
	//and update info structure
	png_read_update_info(png_ptr,info_ptr);

	//-------------------------------------------------------------------------
	//create layer
	last_error=PLUGIN_ERR_INTERNAL;

	data->width=width;
	data->height=height;
	data->meta_data=0;
	if(data->create(data)==0)
	{
		last_error=PLUGIN_ERR_OUT_OF_MEM;
		goto err_internal;
	}

	//insert MARK attachment
	if(png_get_valid(png_ptr,info_ptr,PNG_INFO_pHYs))
	{
		if((data->meta_write)&&(data->pi_MetaMarkup))
		{
			int ppi_x,ppi_y;
			char markup[3072];
			markup[0]=0;
			ppi_x=png_get_x_pixels_per_inch(png_ptr,info_ptr);
			ppi_y=png_get_y_pixels_per_inch(png_ptr,info_ptr);
			if((ppi_x>0)&&(ppi_y>0))
			{
				data->pi_MetaMarkup->ppi_set(markup,3072,ppi_x,ppi_y);
				data->meta_data=(unsigned char*)markup;
				data->meta_write(data,"MARK",strlen(markup)+1);
			}
		}
	}

	//-------------------------------------------------------------------------
	//read the image
	for(row=0;row<height;row++)
	{
		row_pointers[row]=(unsigned char*)(data->lp_pix+(row*data->pitch));
	}
	png_read_image(png_ptr,row_pointers);

	//read rest of file, and get additional chunks in info_ptr
	png_read_end(png_ptr,info_ptr);

	//clean up after the read, and free any memory allocated
	png_destroy_read_struct(&png_ptr,&info_ptr,NULL);

	//-------------------------------------------------------------------------
	//close the file
	fclose(fp);
	data->img_type=PLUGIN_IMGTYPE_LAYERED;
	data->img_subtype=PLUGIN_IMGSUBTYPE_NORMAL;
	return 1;

	//=========================================================================
err_internal:
	png_destroy_read_struct(&png_ptr,&info_ptr,NULL);
	fclose(fp);
	if(wcslen(str_error_list))
	{
		_snwprintf(data->error_hint,PLUGIN_ERROR_HINT_SIZE,L"%s",
			str_error_list);
	}
	return last_error;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/******************************************************************************
This is the plugin DLL function that Chasys Draw IES calls to save images
******************************************************************************/
int flt_SaveImage(flt_IMAGE_S* data)
{
	unsigned int i,j;
	unsigned long ul,*p_mem=0,*p_src,*p_dst;
	png_infop info_ptr;
	png_structp png_ptr;
	png_bytep row_pointers[64*1024];
	FILE *fp;

	//-------------------------------------------------------------------------
	if(!data)
	{
		return PLUGIN_ERR_BAD_HOST;
	}
	if(!data->pi_BasicImage)
	{
		return PLUGIN_ERR_BAD_HOST;
	}
	if(data->pi_BasicImage->version<PI_BASICIMAGE_VERSION)
	{
		return PLUGIN_ERR_BAD_HOST;
	}

	//-------------------------------------------------------------------------
	if(!(flt_CheckSave(wcsrchr(data->file,'.'))))
	{
		return PLUGIN_ERR_NO_SUPPORT;
	}
	if(data->layer_count>1)
	{
		return PLUGIN_ERR_NOT_FLAT;
	}

	//-------------------------------------------------------------------------
	memset(&settings,0,sizeof(plugin_config));
	if(data->pi_StateStore)
	{
		if(data->pi_StateStore->version>=PI_STATESTORE_VERSION)
		{
			data->pi_StateStore->perm_read(
				MY_STORE_NAME,&settings,sizeof(plugin_config));
			if(settings.version!=SETTINGS_VERSION)
			{
				memset(&settings,0,sizeof(plugin_config));
				settings.version=SETTINGS_VERSION;
			}
		}
	}

	//-------------------------------------------------------------------------
	//open the file stream for binary writing
	fp=_wfopen(data->file,L"wb");
	if(!fp)
	{
		return PLUGIN_ERR_FILE_OPEN_W;
	}

	//-------------------------------------------------------------------------
	//create and initialize library structures
	png_ptr=png_create_write_struct(PNG_LIBPNG_VER_STRING,0,0,0);
	if(!png_ptr)
	{
		fclose(fp);
		return PLUGIN_ERR_INTERNAL;
	}

	png_ptr->error_fn=my_png_error;
	str_error_list[0]=0;
	info_ptr=png_create_info_struct(png_ptr);
	if(!info_ptr)
	{
		fclose(fp);
		png_destroy_write_struct(&png_ptr,NULL);
		return PLUGIN_ERR_INTERNAL;
	}

	if(setjmp(png_jmpbuf(png_ptr)))
	{
		goto err_internal;
	}

	//-------------------------------------------------------------------------
	//get the bitmap data, load it into buffer
	ul=data->width*data->height*4;
	p_mem=(unsigned long*)GlobalAlloc(GMEM_FIXED,ul);
	if(!p_mem)
	{
		png_destroy_write_struct(&png_ptr,&info_ptr);
		fclose(fp);
		return PLUGIN_ERR_OUT_OF_MEM;
	}
	for(j=0;j<data->height;j++)
	{
		p_dst=p_mem+(data->width*j);
		p_src=data->lp_pix+(data->pitch*j);
		for(i=0;i<data->width;i++)
		{
			p_dst[i]=
				((p_src[i]>>16)&0x000000ff)|
				( p_src[i]     &0x0000ff00)|
				((p_src[i]<<16)&0x00ff0000)|
				(( ~(p_src[i]))&0xff000000);
		}
		row_pointers[j]=(unsigned char*)(&p_mem[(j*data->width)]);
	}

	//-------------------------------------------------------------------------
	//start file output, initialize image information
	png_init_io(png_ptr,fp);
	png_set_IHDR(png_ptr,info_ptr,data->width,data->height,8,
		PNG_COLOR_TYPE_RGBA,PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_BASE,PNG_FILTER_TYPE_BASE);

	if((data->meta_read)&&(data->pi_MetaMarkup))
	{
		if(data->meta_read(data,"MARK"))
		{
			if(data->meta_data)
			{
				double ppi_x,ppi_y;
				char* markup=(char*)data->meta_data;
				if(data->pi_MetaMarkup->ppi_get(markup,&ppi_x,&ppi_y))
				{
					i=(int)((ppi_x/0.0254)+0.5);
					j=(int)((ppi_y/0.0254)+0.5);
					png_set_pHYs(png_ptr,info_ptr,i,j,PNG_RESOLUTION_METER);
				}
			}
		}
	}

	//png_set_gAMA(png_ptr,info_ptr,gamma);

	png_text text_ptr[1];
	char key0[]="Author";
	char text0[]=SUITE_FULLNAME_A;
	text_ptr[0].key=key0;
	text_ptr[0].text=text0;
	text_ptr[0].compression=PNG_TEXT_COMPRESSION_NONE;
	text_ptr[0].itxt_length=0;
	text_ptr[0].lang=NULL;
	text_ptr[0].lang_key=NULL;
	png_set_text(png_ptr,info_ptr,text_ptr,1);

	//-------------------------------------------------------------------------
	//write the file header
	png_write_info(png_ptr,info_ptr);

	//-------------------------------------------------------------------------
	//write the image data
	//png_set_bgr(png_ptr);//flip BGR pixels to RGB
	png_write_image(png_ptr,row_pointers);
	png_write_end(png_ptr,info_ptr);

	//-------------------------------------------------------------------------
	//clean up, free any memory,close the file
	if(p_mem)
	{
		GlobalFree(p_mem);
		p_mem=0;
	}
	png_destroy_write_struct(&png_ptr,&info_ptr);
	fclose(fp);
	return 1;

	//=========================================================================
err_internal:
	if(p_mem)
	{
		GlobalFree(p_mem);
		p_mem=0;
	}
	png_destroy_write_struct(&png_ptr,&info_ptr);
	fclose(fp);
	if(wcslen(str_error_list))
	{
		_snwprintf(data->error_hint,PLUGIN_ERROR_HINT_SIZE,L"%s",
			str_error_list);
	}
	return PLUGIN_ERR_INTERNAL;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/******************************************************************************
This function processes messages for about dialog
******************************************************************************/
LRESULT CALLBACK msg_dlg_about(HWND hDlg,UINT msg,WPARAM wpar,LPARAM lpar)
{
	wchar_t* pstr;

	//-------------------------------------------------------------------------
	switch(msg)
	{
	case WM_INITDIALOG:
		SetWindowText(hDlg,L"About " PLUGIN_NAME);
		pstr=string_ReadTextResource(api_inst,(LPCTSTR)IDR_README,L"TEXT");
		if(pstr)
		{
			SetDlgItemText(hDlg,IDC_EDIT1,pstr);
			GlobalFree(pstr);
		}
		window_Center(hDlg,GetParent(hDlg));
		return 1|lpar;
	case WM_COMMAND:
		switch(LOWORD(wpar))
		{
		case IDOK    :
			EndDialog(hDlg,1);
			break;
		case IDCANCEL:
			EndDialog(hDlg,0);
			break;
		}
		return 1;
	}
	return 0;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/******************************************************************************
This function processes messages for config dialog
******************************************************************************/
LRESULT CALLBACK msg_dlg_config(HWND hDlg,UINT msg,WPARAM wpar,LPARAM lpar)
{
	switch(msg)
	{
	case WM_INITDIALOG:
		window_Center(hDlg,GetParent(hDlg));
		SetWindowText(hDlg,L"Configure " PLUGIN_NAME);
		CheckDlgButton(hDlg,IDC_CHECK1,settings.always_apply_bKGD);
		return 1|lpar;
	case WM_COMMAND:
		switch(LOWORD(wpar))
		{
		case IDOK:
			settings.version=SETTINGS_VERSION;
			settings.always_apply_bKGD=(IsDlgButtonChecked(hDlg,IDC_CHECK1)!=0);
			EndDialog(hDlg,1);
			break;
		case IDCANCEL:
			EndDialog(hDlg,0);
			break;
		}
		return 1;
	}
	return 0;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////